/*    Copyright (C) 2012 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/


#ifndef _fslvtkIO
#define _fslvtkIO

#include <iostream>

#include <string>
#include <fstream>
#include <stdio.h>
#include <algorithm>

#include "armawrap/newmat.h"
#include "meshclass/meshclass.h"
#include "first_lib/first_newmat_vec.h"


namespace fslvtkio{
  //using namespace mesh;

  class fslvtkIOException : public std::exception{

  public:
    const char* errmesg;
    fslvtkIOException(const char* msg)
    {
      errmesg=msg;
    }

  private:
    virtual const char* what() const throw()
    {
      std::cout<<errmesg<<std::endl;
      return errmesg;
    }
  };

  class fslvtkIO {
  public:
    enum DataType{ POLYDATA, UNSTRUCTURED_GRID };
    fslvtkIO();
    fslvtkIO(const std::string & filename,const DataType i);

    ~fslvtkIO();


    //----------------------SET COMMANDS-------------------------//
    void setPoints(const NEWMAT::Matrix & pts);
    void setPoints(const std::vector<float> & pts);
    void setCells(const std::vector< std::vector<unsigned int> > & c){ Cells=c; }
    void setCells(const std::vector< std::vector<unsigned int> > & c, const std::vector<short> & c_t ){ Cells=c; Cell_Types=c_t; }

    void appendPointsAndPolygons(const NEWMAT::Matrix & pts, const NEWMAT::Matrix & Polygons);


    void setPolygons(const NEWMAT::Matrix& m){ Polygons=m; }
    void setPolygons(const std::vector< std::vector<unsigned int> >& vm);
    void setMesh(const mesh::Mesh & m);
    //void setScalars(const Matrix & sc,const string & name);
    void setScalars(const NEWMAT::Matrix& m){ Scalars=m; }

    void setVectors(const NEWMAT::Matrix & vecs,const std::string &name);
    void setVectors(const NEWMAT::Matrix& m){ Vectors=m; }

    template<class T>
    void setScalars(const std::vector<T> & sc);


    void addPointFieldData(const NEWMAT::Matrix& m,const std::string & name, const std::string & type, const std::string & vtkAttType);
    void addCellFieldData(const NEWMAT::Matrix& m,const std::string & name, const std::string & type, const std::string & vtkAttType);

    void addFieldData(const NEWMAT::Matrix& m,const std::string & name, const std::string & type);

    template< class T >
    void addFieldData(const std::vector<T> & m,const std::string & name, const std::string & type);

    template< class T >
    void addFieldData(const T & m,const std::string & name, const std::string & type);

    void replaceFieldData(const NEWMAT::Matrix& m,const std::string & name);

    void addFieldData(std::vector< std::string >,std::string name);


    void printFieldDataNames() { for (std::vector<std::string>::iterator i=fieldDataNumName.begin(); i!=fieldDataNumName.end();i++) std::cout<<"field "<<*i<<std::endl; }


    //----------------------GET COMMANDS----------------------------//
    NEWMAT::Matrix getPointsAsMatrix() const { return Points; }
    NEWMAT::ColumnVector getPointsAsColumnVector() const ;



    template<class T>
    std::vector<T> getPointsAsVector();

    template<class T>
    std::vector< std::vector<T> > getPointsAsVectorOfVectors(){ return FIRST_LIB::first_newmat_vector::matrixToVectorOfVectors<T>(Points); }

    template<class T>
    std::vector< std::vector<T> > getPolygonsAsVectorOfVectors(){ return FIRST_LIB::first_newmat_vector::matrixToVectorOfVectors<T>(Polygons); }

    NEWMAT::ReturnMatrix getPolygons() const { return Polygons; };
    std::vector< std::vector<unsigned int> > getCells() const { return Cells; }
    std::vector<short> getCellTypes() const { return Cell_Types; }

    std::string getFieldName(const int & ind) const { return fieldDataNumName.at(ind); };
    unsigned int getNumberOfFields() const { return fieldDataNumName.size(); };
    NEWMAT::Matrix getField(const std::string & name);
    NEWMAT::Matrix getField(const std::string & name, unsigned int & ind);
    void setField(const std::string & name, const NEWMAT::Matrix & data);

    NEWMAT::ReturnMatrix getScalars(){ return Scalars; }

    template<class T>
    std::vector<T> getScalars();


    NEWMAT::ReturnMatrix getVectors(){ return Vectors; }


    //----------------------I/O COMMANDS----------------------------//
    template<class T>
    void writePoints(std::ofstream & fshape, const std::string & str_typename);
    void writePolygons(std::ofstream & fshape);
    void writeCells(std::ofstream & fshape);
    //void writeUnstructuredGridCells(ofstream & fshape);
    void writeUnstructuredGridCellTypes(std::ofstream & fshape);
    void save(std::string s);

    void readUnstructuredGrid(std::string fname);
    void readPolyData(std::string fname);
    template<class T>
    void writePointData(std::ofstream & fshape, const std::string & str_typename );

    template <class T>
    void writeNumericField(std::ofstream & fvtk, const std::string & name, const std::string & type, const NEWMAT::Matrix & Data);
    void writeStringField(std::ofstream & fvtk, const std::string & name, const std::vector<std::string> & v_string);

    void readFieldData(std::ifstream & fvtk);
    void readPointData(std::ifstream & fvtk, std::string & nextData);
    bool  readPoints(std::ifstream & fvtk);
    bool  readPolygons(std::ifstream & fvtk);

    template <class T>
    NEWMAT::ReturnMatrix readField(std::ifstream & fvtk, const int & nrows,const int & mcols);

    void displayNumericFieldDataNames();
    void displayNumericField(const std::string & name);

    //----------------------SETTING/ACCESS OF STATE VARIABLES----------------------------//
    void setSwitchRowsCols(bool n) { SWITCH_ROWS_COLS=n; }

    void setMAX(bool b) { MAX_SET=b; }
    void setMAX_Val( unsigned int n ) { MAX=n; }
    void setDataType(const DataType & dtype ){ dt=dtype; }
    void setBinaryWrite(bool state) { BINARY=state; }
    bool getBinaryWrite() { return BINARY; }
    static bool myExceptions(int e);

  protected:
    NEWMAT::Matrix Scalars;
    NEWMAT::Matrix Vectors;
    NEWMAT::Matrix Points;
    //cell data
    NEWMAT::Matrix Polygons;



  private:
#define BINFLAG 42
    //----------------------STATE VARIABLES/CONSTANTS----------------------------//

#ifdef PPC64
	int m_n;
#endif
    bool BINARY;//state variable for read write
    bool SWAP_BYTES;//only used for binary read
    bool MAX_SET; //is a max number of columns imposed
    bool SWITCH_ROWS_COLS;//SWITCHES THE COLUMN/ROWS , USED FOR FIXING OLD FILES

    unsigned int ST_COUNT;

    unsigned int MAX;
    DataType dt;//i.e. POLYDATA




    //----------------------READ COMMANDS----------------------------//

    //----------------------DATA----------------------------//

    //point data
    std::string scalarsName, vectorsName;

    std::vector< std::vector<unsigned int> > Cells;
    std::vector<short> Cell_Types;

    std::string cd_scalarsName, cd_vectorsName;
    NEWMAT::Matrix cd_Scalars;
    NEWMAT::Matrix cd_Vectors;


    //field data
    std::vector< NEWMAT::Matrix > fieldDataNum;
    std::vector< std::string > fieldDataNumName;
    std::vector< std::string > fieldDataNumType;

    std::vector< std::vector<std::string> > fieldDataStr;
    std::vector< std::string > fieldDataStrName;

    //defaults FIELDData
    std::vector< std::string > pd_list;
    std::vector< std::string > pd_type;

    std::vector< std::string > cd_list;
    std::vector< std::string > cd_type;

  };
}
#endif
